from torch import Tensor, ones_like, allclose, cat, rand, manual_seed
from torch.nn import Linear, MSELoss, ReLU, Sequential
from backpack.hessianfree.hvp import hessian_vector_product
from typing import List

x1 = Tensor([1.0])
x2 = Tensor([2.0])

x1.requires_grad = True
x2.requires_grad = True

expected = Tensor([14.0, -32.0])

f = x1 ** 2 + 2 * x1 ** 2 * x2 - 3 * x2 ** 3

parameters = [x1, x2]
v_list = [ones_like(param) for param in parameters]
Hv_list = hessian_vector_product(f, parameters, v_list)

print(Hv_list)

Hv = cat([v.flatten() for v in Hv_list])

if allclose(Hv, expected):
    print("Results match.")
else:
    print("Results don't match.")

def hessian_row_sum(f: Tensor, parameters: List[Tensor]) -> Tensor:
    """Compute the Hessian row-wise sum and return it as 1d tensor.

    Args:
        f: Function of the Hessian.
        parameters: Variables of f. Need to have ``requires_grad=True``.

    Returns:
        Hessian row sum as vector.
    """
    Hv_list = hessian_vector_product(
        f, parameters, [ones_like(param) for param in parameters]
    )

    return cat([v.flatten() for v in Hv_list])

batch_size = 3
D_in = 5
D_out = 4

X = rand(batch_size, D_in)
y = rand(batch_size, D_out)

model = Sequential(Linear(D_in, D_out, bias=False), ReLU())
loss_func = MSELoss()

output = model(X)
loss = loss_func(output, y)

result = hessian_row_sum(loss, list(model.parameters()))

print(result.shape)
